@@ -1,6 +1,7 @@
 /*
  * Copyright (C) 2014 Daniel Vrátil <dvratil@redhat.com>
  * Copyright (c) 2016 Elvis Angelaccio <elvis.angelaccio@kde.org>
+ * Copyright (c) 2019 David Barchiesi <david@barchie.si>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,11 +30,24 @@ public:
     explicit GDriveUrl(const QUrl &url);
 
     QString account() const;
+    QString filename() const;
     bool isRoot() const;
     bool isAccountRoot() const;
+    bool isNewAccountPath() const;
+    bool isTopLevel() const;
+    bool isSharedDrivesRoot() const;
+    bool isSharedDrive() const;
+    bool isTrashDir() const;
+    bool isTrashed() const;
+    QUrl url() const;
     QString parentPath() const;
     QStringList pathComponents() const;
 
+    static const QString Scheme;
+    static const QString SharedDrivesDir;
+    static const QString TrashDir;
+    static const QString NewAccountPath;
+
 private:
     QUrl m_url;
     QStringList m_components;
