@@ -25,33 +25,38 @@
 #include "gdriveversion.h"
 
 #include <QApplication>
-#include <QUrlQuery>
+#include <QMimeDatabase>
+#include <QNetworkReply>
+#include <QNetworkRequest>
 #include <QTemporaryFile>
+#include <QUrlQuery>
+#include <QUuid>
 
 #include <KGAPI/Account>
 #include <KGAPI/AuthJob>
 #include <KGAPI/Drive/About>
 #include <KGAPI/Drive/AboutFetchJob>
 #include <KGAPI/Drive/ChildReference>
-#include <KGAPI/Drive/ChildReferenceFetchJob>
 #include <KGAPI/Drive/ChildReferenceCreateJob>
+#include <KGAPI/Drive/ChildReferenceFetchJob>
+#include <KGAPI/Drive/Drives>
+#include <KGAPI/Drive/DrivesCreateJob>
+#include <KGAPI/Drive/DrivesDeleteJob>
+#include <KGAPI/Drive/DrivesFetchJob>
+#include <KGAPI/Drive/DrivesModifyJob>
 #include <KGAPI/Drive/File>
 #include <KGAPI/Drive/FileCopyJob>
 #include <KGAPI/Drive/FileCreateJob>
-#include <KGAPI/Drive/FileModifyJob>
-#include <KGAPI/Drive/FileTrashJob>
-#include <KGAPI/Drive/FileFetchJob>
 #include <KGAPI/Drive/FileFetchContentJob>
+#include <KGAPI/Drive/FileFetchJob>
+#include <KGAPI/Drive/FileModifyJob>
 #include <KGAPI/Drive/FileSearchQuery>
+#include <KGAPI/Drive/FileTrashJob>
 #include <KGAPI/Drive/ParentReference>
 #include <KGAPI/Drive/Permission>
 #include <KIO/AccessManager>
 #include <KIO/Job>
 #include <KLocalizedString>
-#include <KPim/kgapi_version.h>
-
-#include <QNetworkRequest>
-#include <QNetworkReply>
 
 using namespace KGAPI2;
 using namespace Drive;
@@ -98,7 +103,7 @@ KIOGDrive::~KIOGDrive()
 
 KIOGDrive::Action KIOGDrive::handleError(const KGAPI2::Job &job, const QUrl &url)
 {
-    qCDebug(GDRIVE) << "Job status code:" << job.error() << "- message:" << job.errorString();
+    qCDebug(GDRIVE) << "Completed job" << (&job) << "error code:" << job.error() << "- message:" << job.errorString();
 
     switch (job.error()) {
         case KGAPI2::OK:
@@ -140,31 +145,34 @@ KIOGDrive::Action KIOGDrive::handleError(const KGAPI2::Job &job, const QUrl &url
 void KIOGDrive::fileSystemFreeSpace(const QUrl &url)
 {
     const auto gdriveUrl = GDriveUrl(url);
-    const QString accountId = gdriveUrl.account();
-    if (accountId == QLatin1String("new-account")) {
+    if (gdriveUrl.isNewAccountPath()) {
+        qCDebug(GDRIVE) << "fileSystemFreeSpace is not supported for new-account url";
         finished();
         return;
     }
-    if (!gdriveUrl.isRoot()) {
-        AboutFetchJob aboutFetch(getAccount(accountId));
-#if KGAPI_VERSION >= QT_VERSION_CHECK(5, 12, 0)
-        aboutFetch.setFields({
-            About::Fields::Kind,
-            About::Fields::QuotaBytesTotal,
-            About::Fields::QuotaBytesUsedAggregate
-        });
-#endif
-        if (runJob(aboutFetch, url, accountId)) {
-            const AboutPtr about = aboutFetch.aboutData();
-            if (about) {
-                setMetaData(QStringLiteral("total"), QString::number(about->quotaBytesTotal()));
-                setMetaData(QStringLiteral("available"), QString::number(about->quotaBytesTotal() - about->quotaBytesUsedAggregate()));
-                finished();
-                return;
-            }
+    if (gdriveUrl.isRoot()) {
+        qCDebug(GDRIVE) << "fileSystemFreeSpace is not supported for gdrive root url";
+        error(KIO::ERR_CANNOT_STAT, url.toDisplayString());
+        return;
+    }
+
+    qCDebug(GDRIVE) << "Getting fileSystemFreeSpace for" << url;
+    const QString accountId = gdriveUrl.account();
+    AboutFetchJob aboutFetch(getAccount(accountId));
+    aboutFetch.setFields({
+        About::Fields::Kind,
+        About::Fields::QuotaBytesTotal,
+        About::Fields::QuotaBytesUsedAggregate
+    });
+    if (runJob(aboutFetch, url, accountId)) {
+        const AboutPtr about = aboutFetch.aboutData();
+        if (about) {
+            setMetaData(QStringLiteral("total"), QString::number(about->quotaBytesTotal()));
+            setMetaData(QStringLiteral("available"), QString::number(about->quotaBytesTotal() - about->quotaBytesUsedAggregate()));
+            finished();
+            return;
         }
     }
-    error(KIO::ERR_CANNOT_STAT, url.toDisplayString());
 }
 
 AccountPtr KIOGDrive::getAccount(const QString &accountName)
@@ -195,46 +203,48 @@ KIO::UDSEntry KIOGDrive::fileToUDSEntry(const FilePtr &origFile, const QString &
         GDriveHelper::convertFromGDocs(file);
     }
 
-    entry.insert(KIO::UDSEntry::UDS_NAME, file->title());
-    entry.insert(KIO::UDSEntry::UDS_DISPLAY_NAME, file->title());
-    entry.insert(KIO::UDSEntry::UDS_COMMENT, file->description());
+    entry.fastInsert(KIO::UDSEntry::UDS_NAME, file->title());
+    entry.fastInsert(KIO::UDSEntry::UDS_DISPLAY_NAME, file->title());
+    entry.fastInsert(KIO::UDSEntry::UDS_COMMENT, file->description());
 
     if (file->isFolder()) {
-        entry.insert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
-        entry.insert(KIO::UDSEntry::UDS_SIZE, 0);
+        entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
+        entry.fastInsert(KIO::UDSEntry::UDS_SIZE, 0);
         isFolder = true;
     } else {
-        entry.insert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFREG);
-        entry.insert(KIO::UDSEntry::UDS_MIME_TYPE, file->mimeType());
-        entry.insert(KIO::UDSEntry::UDS_SIZE, file->fileSize());
+        entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFREG);
+        entry.fastInsert(KIO::UDSEntry::UDS_MIME_TYPE, file->mimeType());
+        entry.fastInsert(KIO::UDSEntry::UDS_SIZE, file->fileSize());
 
-        entry.insert(KIO::UDSEntry::UDS_URL, fileToUrl(origFile, path).toString());
+        entry.fastInsert(KIO::UDSEntry::UDS_URL, fileToUrl(origFile, path).toString());
     }
 
-    entry.insert(KIO::UDSEntry::UDS_CREATION_TIME, file->createdDate().toTime_t());
-    entry.insert(KIO::UDSEntry::UDS_MODIFICATION_TIME, file->modifiedDate().toTime_t());
-    entry.insert(KIO::UDSEntry::UDS_ACCESS_TIME, file->lastViewedByMeDate().toTime_t());
+    entry.fastInsert(KIO::UDSEntry::UDS_CREATION_TIME, file->createdDate().toTime_t());
+    entry.fastInsert(KIO::UDSEntry::UDS_MODIFICATION_TIME, file->modifiedDate().toTime_t());
+    entry.fastInsert(KIO::UDSEntry::UDS_ACCESS_TIME, file->lastViewedByMeDate().toTime_t());
     if (!file->ownerNames().isEmpty()) {
-        entry.insert(KIO::UDSEntry::UDS_USER, file->ownerNames().first());
+        entry.fastInsert(KIO::UDSEntry::UDS_USER, file->ownerNames().first());
     }
 
     if (!isFolder) {
         if (file->editable()) {
-            entry.insert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
+            entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
         } else {
-            entry.insert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR | S_IRGRP | S_IROTH);
+            entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR | S_IRGRP | S_IROTH);
         }
     } else {
-        entry.insert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP | S_IROTH | S_IXOTH);
+        entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP | S_IROTH | S_IXOTH);
     }
 
+    entry.fastInsert(KIO::UDSEntry::UDS_EXTRA, file->alternateLink().toString());
+
     return entry;
 }
 
 QUrl KIOGDrive::fileToUrl(const FilePtr &file, const QString &path) const
 {
     QUrl url;
-    url.setScheme(QStringLiteral("gdrive"));
+    url.setScheme(GDriveUrl::Scheme);
     url.setPath(path + QLatin1Char('/') + file->title());
 
     QUrlQuery urlQuery;
@@ -249,16 +259,50 @@ void KIOGDrive::openConnection()
     qCDebug(GDRIVE) << "Ready to talk to GDrive";
 }
 
+KIO::UDSEntry KIOGDrive::newAccountUDSEntry()
+{
+    KIO::UDSEntry entry;
+
+    entry.fastInsert(KIO::UDSEntry::UDS_NAME, GDriveUrl::NewAccountPath);
+    entry.fastInsert(KIO::UDSEntry::UDS_DISPLAY_NAME, i18nc("login in a new google account", "New account"));
+    entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
+    entry.fastInsert(KIO::UDSEntry::UDS_ICON_NAME, QStringLiteral("list-add-user"));
+    entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR);
+
+    return entry;
+}
+
 KIO::UDSEntry KIOGDrive::accountToUDSEntry(const QString &accountNAme)
 {
     KIO::UDSEntry entry;
 
-    entry.insert(KIO::UDSEntry::UDS_NAME, accountNAme);
-    entry.insert(KIO::UDSEntry::UDS_DISPLAY_NAME, accountNAme);
-    entry.insert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
-    entry.insert(KIO::UDSEntry::UDS_SIZE, 0);
-    entry.insert(KIO::UDSEntry::UDS_ACCESS, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
-    entry.insert(KIO::UDSEntry::UDS_ICON_NAME, QStringLiteral("folder-gdrive"));
+    entry.fastInsert(KIO::UDSEntry::UDS_NAME, accountNAme);
+    entry.fastInsert(KIO::UDSEntry::UDS_DISPLAY_NAME, accountNAme);
+    entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
+    entry.fastInsert(KIO::UDSEntry::UDS_SIZE, 0);
+    entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
+    entry.fastInsert(KIO::UDSEntry::UDS_ICON_NAME, QStringLiteral("folder-gdrive"));
+
+    return entry;
+}
+
+KIO::UDSEntry KIOGDrive::sharedDriveToUDSEntry(const DrivesPtr &sharedDrive)
+{
+    KIO::UDSEntry entry;
+
+    qlonglong udsAccess = S_IRUSR | S_IXUSR | S_IRGRP;
+    if (sharedDrive->capabilities()->canRenameDrive() || sharedDrive->capabilities()->canDeleteDrive()) {
+        udsAccess |= S_IWUSR;
+    }
+
+    entry.fastInsert(KIO::UDSEntry::UDS_NAME, sharedDrive->id());
+    entry.fastInsert(KIO::UDSEntry::UDS_DISPLAY_NAME, sharedDrive->name());
+    entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
+    entry.fastInsert(KIO::UDSEntry::UDS_SIZE, 0);
+    entry.fastInsert(KIO::UDSEntry::UDS_CREATION_TIME, sharedDrive->createdDate().toSecsSinceEpoch());
+    entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, udsAccess);
+    entry.fastInsert(KIO::UDSEntry::UDS_HIDDEN, sharedDrive->hidden());
+    entry.fastInsert(KIO::UDSEntry::UDS_ICON_NAME, QStringLiteral("folder-gdrive"));
 
     return entry;
 }
@@ -295,25 +339,140 @@ void KIOGDrive::listAccounts()
         const KIO::UDSEntry entry = accountToUDSEntry(account);
         listEntry(entry);
     }
-    KIO::UDSEntry newAccountEntry;
-    newAccountEntry.insert(KIO::UDSEntry::UDS_NAME, QStringLiteral("new-account"));
-    newAccountEntry.insert(KIO::UDSEntry::UDS_DISPLAY_NAME, i18nc("login in a new google account", "New account"));
-    newAccountEntry.insert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
-    newAccountEntry.insert(KIO::UDSEntry::UDS_ICON_NAME, QStringLiteral("list-add-user"));
+
+    KIO::UDSEntry newAccountEntry = newAccountUDSEntry();
     listEntry(newAccountEntry);
 
     // Create also non-writable UDSentry for "."
     KIO::UDSEntry entry;
-    entry.insert(KIO::UDSEntry::UDS_NAME, QStringLiteral("."));
-    entry.insert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
-    entry.insert(KIO::UDSEntry::UDS_SIZE, 0);
-    entry.insert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
+    entry.fastInsert(KIO::UDSEntry::UDS_NAME, QStringLiteral("."));
+    entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
+    entry.fastInsert(KIO::UDSEntry::UDS_SIZE, 0);
+    entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
     listEntry(entry);
 
     finished();
     return;
 }
 
+void KIOGDrive::listSharedDrivesRoot(const QUrl &url)
+{
+    const auto gdriveUrl = GDriveUrl(url);
+    const QString accountId = gdriveUrl.account();
+    DrivesFetchJob sharedDrivesFetchJob(getAccount(accountId));
+    sharedDrivesFetchJob.setFields({
+        Drives::Fields::Kind,
+        Drives::Fields::Id,
+        Drives::Fields::Name,
+        Drives::Fields::Hidden,
+        Drives::Fields::CreatedDate,
+        Drives::Fields::Capabilities
+    });
+
+    if (runJob(sharedDrivesFetchJob, url, accountId)) {
+        const auto objects = sharedDrivesFetchJob.items();
+        for (const auto &object : objects) {
+            const DrivesPtr sharedDrive = object.dynamicCast<Drives>();
+            const KIO::UDSEntry entry = sharedDriveToUDSEntry(sharedDrive);
+            listEntry(entry);
+        }
+
+        auto entry = fetchSharedDrivesRootEntry(accountId, FetchEntryFlags::CurrentDir);
+        listEntry(entry);
+
+        finished();
+    }
+}
+
+bool KIOGDrive::createSharedDrive(const QUrl &url)
+{
+    const auto gdriveUrl = GDriveUrl(url);
+    const QString accountId = gdriveUrl.account();
+
+    DrivesPtr sharedDrive = DrivesPtr::create();
+    sharedDrive->setName(gdriveUrl.filename());
+
+    const QString requestId = QUuid::createUuid().toString();
+    DrivesCreateJob createJob(requestId, sharedDrive, getAccount(accountId));
+    return runJob(createJob, url, accountId);
+}
+
+bool KIOGDrive::deleteSharedDrive(const QUrl &url)
+{
+    const auto gdriveUrl = GDriveUrl(url);
+    const QString accountId = gdriveUrl.account();
+    DrivesDeleteJob sharedDriveDeleteJob(gdriveUrl.filename(), getAccount(accountId));
+    return runJob(sharedDriveDeleteJob, url, accountId);
+}
+
+void KIOGDrive::statSharedDrive(const QUrl &url)
+{
+    const auto gdriveUrl = GDriveUrl(url);
+    const QString accountId = gdriveUrl.account();
+    DrivesFetchJob sharedDriveFetchJob(gdriveUrl.filename(), getAccount(accountId));
+    sharedDriveFetchJob.setFields({
+        Drives::Fields::Kind,
+        Drives::Fields::Id,
+        Drives::Fields::Name,
+        Drives::Fields::Hidden,
+        Drives::Fields::CreatedDate,
+        Drives::Fields::Capabilities
+    });
+    if (!runJob(sharedDriveFetchJob, url, accountId)) {
+        return;
+    }
+
+    ObjectPtr object = sharedDriveFetchJob.items().at(0);
+    const DrivesPtr sharedDrive = object.dynamicCast<Drives>();
+    const auto entry = sharedDriveToUDSEntry(sharedDrive);
+    statEntry(entry);
+    finished();
+}
+
+KIO::UDSEntry KIOGDrive::fetchSharedDrivesRootEntry(const QString &accountId, FetchEntryFlags flags)
+{
+    // Not every user is allowed to create shared Drives,
+    // check with About resource.
+    bool canCreateDrives = false;
+    AboutFetchJob aboutFetch(getAccount(accountId));
+    aboutFetch.setFields({
+        About::Fields::Kind,
+        About::Fields::CanCreateDrives
+    });
+    QEventLoop eventLoop;
+    QObject::connect(&aboutFetch, &KGAPI2::Job::finished,
+                     &eventLoop, &QEventLoop::quit);
+    eventLoop.exec();
+    if (aboutFetch.error() == KGAPI2::OK || aboutFetch.error() == KGAPI2::NoError) {
+        const AboutPtr about = aboutFetch.aboutData();
+        if (about) {
+            canCreateDrives = about->canCreateDrives();
+        }
+    }
+    qCDebug(GDRIVE) << "Account" << accountId << (canCreateDrives ? "can" : "can't") << "create Drives";
+
+    KIO::UDSEntry entry;
+
+    if (flags == FetchEntryFlags::CurrentDir) {
+        entry.fastInsert(KIO::UDSEntry::UDS_NAME, QStringLiteral("."));
+    } else {
+        entry.fastInsert(KIO::UDSEntry::UDS_NAME, GDriveUrl::SharedDrivesDir);
+        entry.fastInsert(KIO::UDSEntry::UDS_DISPLAY_NAME, i18n("Shared Drives"));
+    }
+    entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
+    entry.fastInsert(KIO::UDSEntry::UDS_SIZE, 0);
+    entry.fastInsert(KIO::UDSEntry::UDS_ICON_NAME, QStringLiteral("folder-gdrive"));
+
+    qlonglong udsAccess = S_IRUSR | S_IXUSR;
+    // If user is allowed to create shared Drives, add write bit on directory
+    if (canCreateDrives) {
+        udsAccess |= S_IWUSR;
+    }
+    entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, udsAccess);
+
+    return entry;
+}
+
 class RecursionDepthCounter
 {
 public:
@@ -343,7 +502,7 @@ int RecursionDepthCounter::sDepth = 0;
 
 QString KIOGDrive::resolveFileIdFromPath(const QString &path, PathFlags flags)
 {
-    qCDebug(GDRIVE) << Q_FUNC_INFO << path;
+    qCDebug(GDRIVE) << "Resolving file ID for" << path;
 
     if (path.isEmpty()) {
         return QString();
@@ -356,15 +515,24 @@ QString KIOGDrive::resolveFileIdFromPath(const QString &path, PathFlags flags)
     }
 
     QUrl url;
-    url.setScheme(QStringLiteral("gdrive"));
+    url.setScheme(GDriveUrl::Scheme);
     url.setPath(path);
     const auto gdriveUrl = GDriveUrl(url);
     Q_ASSERT(!gdriveUrl.isRoot());
 
-    const QStringList components = gdriveUrl.pathComponents();
-    if (gdriveUrl.isAccountRoot() || (components.size() == 2 && components[1] == QLatin1String("trash"))) {
+    if (gdriveUrl.isAccountRoot() || gdriveUrl.isTrashDir()) {
         qCDebug(GDRIVE) << "Resolved" << path << "to \"root\"";
-        return rootFolderId(components[0]);
+        return rootFolderId(gdriveUrl.account());
+    }
+
+    if (gdriveUrl.isSharedDrive()) {
+        qCDebug(GDRIVE) << "Resolved" << path << "to Shared Drive" << gdriveUrl.filename();
+        return gdriveUrl.filename();
+    }
+
+    if (gdriveUrl.isSharedDrivesRoot()) {
+        qCDebug(GDRIVE) << "Resolved" << path << "to Shared Drives root";
+        return QString();
     }
 
     // Try to recursively resolve ID of parent path - either from cache, or by
@@ -381,24 +549,20 @@ QString KIOGDrive::resolveFileIdFromPath(const QString &path, PathFlags flags)
                        (flags & KIOGDrive::PathIsFolder ? FileSearchQuery::Equals : FileSearchQuery::NotEquals),
                        GDriveHelper::folderMimeType());
     }
-    query.addQuery(FileSearchQuery::Title, FileSearchQuery::Equals, components.last());
+    query.addQuery(FileSearchQuery::Title, FileSearchQuery::Equals, gdriveUrl.filename());
     query.addQuery(FileSearchQuery::Parents, FileSearchQuery::In, parentId);
-    query.addQuery(FileSearchQuery::Trashed, FileSearchQuery::Equals, components[1] == QLatin1String("trash"));
+    query.addQuery(FileSearchQuery::Trashed, FileSearchQuery::Equals, gdriveUrl.isTrashed());
 
     const QString accountId = gdriveUrl.account();
     FileFetchJob fetchJob(query, getAccount(accountId));
-#if KGAPI_VERSION >= QT_VERSION_CHECK(5, 12, 0)
     fetchJob.setFields({File::Fields::Id, File::Fields::Title, File::Fields::Labels});
-#else
-    fetchJob.setFields(FileFetchJob::Id | FileFetchJob::Title | FileFetchJob::Labels);
-#endif
     if (!runJob(fetchJob, url, accountId)) {
         return QString();
     }
 
     const ObjectsList objects = fetchJob.items();
     qCDebug(GDRIVE) << objects;
-    if (objects.count() == 0) {
+    if (objects.isEmpty()) {
         qCWarning(GDRIVE) << "Failed to resolve" << path;
         return QString();
     }
@@ -415,10 +579,9 @@ QString KIOGDrive::rootFolderId(const QString &accountId)
 {
     auto it = m_rootIds.constFind(accountId);
     if (it == m_rootIds.cend()) {
+        qCDebug(GDRIVE) << "Getting root ID for" << accountId;
         AboutFetchJob aboutFetch(getAccount(accountId));
-#if KGAPI_VERSION >= QT_VERSION_CHECK(5, 12, 0)
         aboutFetch.setFields({About::Fields::Kind, About::Fields::RootFolderId});
-#endif
         QUrl url;
         if (!runJob(aboutFetch, url, accountId)) {
             return QString();
@@ -442,18 +605,23 @@ void KIOGDrive::listDir(const QUrl &url)
     qCDebug(GDRIVE) << "Going to list" << url;
 
     const auto gdriveUrl = GDriveUrl(url);
-    const QString accountId = gdriveUrl.account();
-    if (accountId == QLatin1String("new-account")) {
+    if (gdriveUrl.isNewAccountPath()) {
         createAccount();
         return;
     }
 
+    const QString accountId = gdriveUrl.account();
     QString folderId;
     if (gdriveUrl.isRoot())  {
         listAccounts();
         return;
     } else if (gdriveUrl.isAccountRoot()) {
+        auto entry = fetchSharedDrivesRootEntry(accountId);
+        listEntry(entry);
         folderId = rootFolderId(accountId);
+    } else if (gdriveUrl.isSharedDrivesRoot()) {
+        listSharedDrivesRoot(url);
+        return;
     } else {
         folderId = m_cache.idForPath(url.path());
         if (folderId.isEmpty()) {
@@ -470,20 +638,16 @@ void KIOGDrive::listDir(const QUrl &url)
     query.addQuery(FileSearchQuery::Trashed, FileSearchQuery::Equals, false);
     query.addQuery(FileSearchQuery::Parents, FileSearchQuery::In, folderId);
     FileFetchJob fileFetchJob(query, getAccount(accountId));
-#if KGAPI_VERSION >= QT_VERSION_CHECK(5, 12, 0)
     const auto extraFields =
-            QStringList({ KGAPI2::Drive::File::Fields::Labels,
-                          KGAPI2::Drive::File::Fields::ExportLinks,
-                          KGAPI2::Drive::File::Fields::LastViewedByMeDate,
-            });
+        QStringList({ KGAPI2::Drive::File::Fields::Labels,
+                      KGAPI2::Drive::File::Fields::ExportLinks,
+                      KGAPI2::Drive::File::Fields::LastViewedByMeDate,
+                      KGAPI2::Drive::File::Fields::AlternateLink,
+        });
     fileFetchJob.setFields(KGAPI2::Drive::FileFetchJob::FieldShorthands::BasicFields + extraFields);
-#else
-    fileFetchJob.setFields((FileFetchJob::BasicFields & ~FileFetchJob::Permissions)
-                            | FileFetchJob::Labels
-                            | FileFetchJob::ExportLinks
-                            | FileFetchJob::LastViewedByMeDate);
-#endif
-    runJob(fileFetchJob, url, accountId);
+    if (!runJob(fileFetchJob, url, accountId)) {
+        return;
+    }
 
     ObjectsList objects = fileFetchJob.items();
     Q_FOREACH (const ObjectPtr &object, objects) {
@@ -498,16 +662,15 @@ void KIOGDrive::listDir(const QUrl &url)
 
     // We also need a non-null and writable UDSentry for "."
     KIO::UDSEntry entry;
-    entry.insert(KIO::UDSEntry::UDS_NAME, QStringLiteral("."));
-    entry.insert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
-    entry.insert(KIO::UDSEntry::UDS_SIZE, 0);
-    entry.insert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP | S_IROTH | S_IXOTH);
+    entry.fastInsert(KIO::UDSEntry::UDS_NAME, QStringLiteral("."));
+    entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
+    entry.fastInsert(KIO::UDSEntry::UDS_SIZE, 0);
+    entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP | S_IROTH | S_IXOTH);
     listEntry(entry);
 
     finished();
 }
 
-
 void KIOGDrive::mkdir(const QUrl &url, int permissions)
 {
     // NOTE: We deliberately ignore the permissions field here, because GDrive
@@ -524,9 +687,18 @@ void KIOGDrive::mkdir(const QUrl &url, int permissions)
         error(KIO::ERR_DOES_NOT_EXIST, url.path());
         return;
     }
+
+    if (gdriveUrl.isSharedDrive()) {
+        qCDebug(GDRIVE) << "Directory is shared drive, creating that instead" << url;
+        if (createSharedDrive(url)) {
+            // runJob will have shown an error otherwise
+            finished();
+        }
+        return;
+    }
+
     QString parentId;
-    const auto components = gdriveUrl.pathComponents();
-    if (components.size() == 2) {
+    if (gdriveUrl.isTopLevel()) {
         parentId = rootFolderId(accountId);
     } else {
         parentId = resolveFileIdFromPath(gdriveUrl.parentPath(), KIOGDrive::PathIsFolder);
@@ -537,19 +709,17 @@ void KIOGDrive::mkdir(const QUrl &url, int permissions)
         return;
     }
 
-    const QString folderName = components.last();
-
     FilePtr file(new File());
-    file->setTitle(folderName);
+    file->setTitle(gdriveUrl.filename());
     file->setMimeType(File::folderMimeType());
 
     ParentReferencePtr parent(new ParentReference(parentId));
     file->setParents(ParentReferencesList() << parent);
 
     FileCreateJob createJob(file, getAccount(accountId));
-    runJob(createJob, url, accountId);
-
-    finished();
+    if (runJob(createJob, url, accountId)) {
+        finished();
+    }
 }
 
 void KIOGDrive::stat(const QUrl &url)
@@ -564,11 +734,31 @@ void KIOGDrive::stat(const QUrl &url)
         return;
     }
     if (gdriveUrl.isAccountRoot()) {
+        qCDebug(GDRIVE) << "stat()ing account root";
         const KIO::UDSEntry entry = accountToUDSEntry(accountId);
         statEntry(entry);
         finished();
         return;
     }
+    if (gdriveUrl.isNewAccountPath()) {
+        qCDebug(GDRIVE) << "stat()ing new-account path";
+        const KIO::UDSEntry entry = newAccountUDSEntry();
+        statEntry(entry);
+        finished();
+        return;
+    }
+    if (gdriveUrl.isSharedDrivesRoot()) {
+        qCDebug(GDRIVE) << "stat()ing Shared Drives root";
+        const auto entry = fetchSharedDrivesRootEntry(accountId);
+        statEntry(entry);
+        finished();
+        return;
+    }
+    if (gdriveUrl.isSharedDrive()) {
+        qCDebug(GDRIVE) << "stat()ing Shared Drive" << url;
+        statSharedDrive(url);
+        return;
+    }
 
     const QUrlQuery urlQuery(url);
     const QString fileId
@@ -582,7 +772,10 @@ void KIOGDrive::stat(const QUrl &url)
     }
 
     FileFetchJob fileFetchJob(fileId, getAccount(accountId));
-    runJob(fileFetchJob, url, accountId);
+    if (!runJob(fileFetchJob, url, accountId)) {
+        qCDebug(GDRIVE) << "Failed stat()ing file" << fileFetchJob.errorString();
+        return;
+    }
 
     const ObjectsList objects = fileFetchJob.items();
     if (objects.count() != 1) {
@@ -631,15 +824,10 @@ void KIOGDrive::get(const QUrl &url)
     }
 
     FileFetchJob fileFetchJob(fileId, getAccount(accountId));
-#if KGAPI_VERSION >= QT_VERSION_CHECK(5, 12, 0)
     fileFetchJob.setFields({File::Fields::Id, File::Fields::MimeType, File::Fields::ExportLinks, File::Fields::DownloadUrl});
-#else
-    fileFetchJob.setFields(FileFetchJob::Id
-                            | FileFetchJob::MimeType
-                            | FileFetchJob::ExportLinks
-                            | FileFetchJob::DownloadUrl);
-#endif
-    runJob(fileFetchJob, url, accountId);
+    if (!runJob(fileFetchJob, url, accountId)) {
+        return;
+    }
 
     const ObjectsList objects = fileFetchJob.items();
     if (objects.count() != 1) {
@@ -658,7 +846,9 @@ void KIOGDrive::get(const QUrl &url)
     mimeType(file->mimeType());
 
     FileFetchContentJob contentJob(downloadUrl, getAccount(accountId));
-    runJob(contentJob, url, accountId);
+    if (!runJob(contentJob, url, accountId)) {
+        return;
+    }
 
     QByteArray contentData = contentJob.data();
 
@@ -676,7 +866,7 @@ void KIOGDrive::get(const QUrl &url)
     finished();
 }
 
-bool KIOGDrive::readPutData(QTemporaryFile &tempFile)
+bool KIOGDrive::readPutData(QTemporaryFile &tempFile, FilePtr &fileMetaData)
 {
     // TODO: Instead of using a temp file, upload directly the raw data (requires
     // support in LibKGAPI)
@@ -704,6 +894,10 @@ bool KIOGDrive::readPutData(QTemporaryFile &tempFile)
             }
         }
     } while (result > 0);
+
+    const QMimeType mime = QMimeDatabase().mimeTypeForFileNameAndData(fileMetaData->title(), &tempFile);
+    fileMetaData->setMimeType(mime.name());
+
     tempFile.close();
 
     if (result == -1) {
@@ -719,7 +913,7 @@ bool KIOGDrive::runJob(KGAPI2::Job &job, const QUrl &url, const QString &account
 {
     KIOGDrive::Action action = KIOGDrive::Fail;
     Q_FOREVER {
-        qCDebug(GDRIVE) << "Running job" << (&job) << "with accessToken" << job.account()->accessToken();
+        qCDebug(GDRIVE) << "Running job" << (&job) << "with accessToken" << GDriveHelper::elideToken(job.account()->accessToken());
         QEventLoop eventLoop;
         QObject::connect(&job, &KGAPI2::Job::finished,
                          &eventLoop, &QEventLoop::quit);
@@ -756,9 +950,10 @@ bool KIOGDrive::putUpdate(const QUrl &url)
         return false;
     }
 
-    const FilePtr file = objects[0].dynamicCast<File>();
+    FilePtr file = objects[0].dynamicCast<File>();
+
     QTemporaryFile tmpFile;
-    if (!readPutData(tmpFile)) {
+    if (!readPutData(tmpFile, file)) {
         error(KIO::ERR_CANNOT_READ, url.path());
         return false;
     }
@@ -782,10 +977,9 @@ bool KIOGDrive::putCreate(const QUrl &url)
         error(KIO::ERR_ACCESS_DENIED, url.path());
         return false;
     }
-    const auto components = gdriveUrl.pathComponents();
-    if (components.length() == 2) {
-        // Creating in root directory
-    } else {
+
+    if (!gdriveUrl.isTopLevel()) {
+        // Not creating in root directory, fill parent references
         const QString parentId = resolveFileIdFromPath(gdriveUrl.parentPath());
         if (parentId.isEmpty()) {
             error(KIO::ERR_DOES_NOT_EXIST, url.adjusted(QUrl::RemoveFilename|QUrl::StripTrailingSlash).path());
@@ -795,7 +989,7 @@ bool KIOGDrive::putCreate(const QUrl &url)
     }
 
     FilePtr file(new File);
-    file->setTitle(components.last());
+    file->setTitle(gdriveUrl.filename());
     file->setParents(parentReferences);
     /*
     if (hasMetaData(QLatin1String("modified"))) {
@@ -806,7 +1000,7 @@ bool KIOGDrive::putCreate(const QUrl &url)
     */
 
     QTemporaryFile tmpFile;
-    if (!readPutData(tmpFile)) {
+    if (!readPutData(tmpFile, file)) {
         error(KIO::ERR_CANNOT_READ, url.path());
         return false;
     }
@@ -831,6 +1025,14 @@ void KIOGDrive::put(const QUrl &url, int permissions, KIO::JobFlags flags)
 
     qCDebug(GDRIVE) << Q_FUNC_INFO << url;
 
+    const auto gdriveUrl = GDriveUrl(url);
+
+    if (gdriveUrl.isSharedDrive()) {
+        qCDebug(GDRIVE) << "Can't create files in Shared Drives root" << url;
+        error(KIO::ERR_CANNOT_WRITE, url.path());
+        return;
+    }
+
     if (QUrlQuery(url).hasQueryItem(QStringLiteral("id"))) {
         if (!putUpdate(url)) {
             return;
@@ -894,13 +1096,10 @@ void KIOGDrive::copy(const QUrl &src, const QUrl &dest, int permissions, KIO::Jo
         return;
     }
     FileFetchJob sourceFileFetchJob(sourceFileId, getAccount(sourceAccountId));
-#if KGAPI_VERSION >= QT_VERSION_CHECK(5, 12, 0)
     sourceFileFetchJob.setFields({File::Fields::Id, File::Fields::ModifiedDate, File::Fields::LastViewedByMeDate, File::Fields::Description});
-#else
-    sourceFileFetchJob.setFields(FileFetchJob::Id | FileFetchJob::ModifiedDate |
-                                 FileFetchJob::LastViewedByMeDate | FileFetchJob::Description);
-#endif
-    runJob(sourceFileFetchJob, src, sourceAccountId);
+    if (!runJob(sourceFileFetchJob, src, sourceAccountId)) {
+        return;
+    }
 
     const ObjectsList objects = sourceFileFetchJob.items();
     if (objects.count() != 1) {
@@ -917,9 +1116,7 @@ void KIOGDrive::copy(const QUrl &src, const QUrl &dest, int permissions, KIO::Jo
     }
 
     QString destDirId;
-    const auto destPathComps = destGDriveUrl.pathComponents();
-    const QString destFileName = destPathComps.last();
-    if (destPathComps.size() == 2) {
+    if (destGDriveUrl.isTopLevel()) {
         destDirId = rootFolderId(destAccountId);
     } else {
         destDirId = resolveFileIdFromPath(destGDriveUrl.parentPath(), KIOGDrive::PathIsFolder);
@@ -927,16 +1124,16 @@ void KIOGDrive::copy(const QUrl &src, const QUrl &dest, int permissions, KIO::Jo
     destParentReferences << ParentReferencePtr(new ParentReference(destDirId));
 
     FilePtr destFile(new File);
-    destFile->setTitle(destFileName);
+    destFile->setTitle(destGDriveUrl.filename());
     destFile->setModifiedDate(sourceFile->modifiedDate());
     destFile->setLastViewedByMeDate(sourceFile->lastViewedByMeDate());
     destFile->setDescription(sourceFile->description());
     destFile->setParents(destParentReferences);
 
     FileCopyJob copyJob(sourceFile, destFile, getAccount(sourceAccountId));
-    runJob(copyJob, dest, sourceAccountId);
-
-    finished();
+    if (runJob(copyJob, dest, sourceAccountId)) {
+        finished();
+    }
 }
 
 void KIOGDrive::del(const QUrl &url, bool isfile)
@@ -953,6 +1150,15 @@ void KIOGDrive::del(const QUrl &url, bool isfile)
     // their local trashes. This however requires fixing the first FIXME first,
     // otherwise we are risking severe data loss.
 
+    const auto gdriveUrl = GDriveUrl(url);
+
+    // Trying to delete the Team Drive root is pointless
+    if (gdriveUrl.isSharedDrivesRoot()) {
+        qCDebug(GDRIVE) << "Tried deleting Shared Drives root.";
+        error(KIO::ERR_SLAVE_DEFINED, i18n("Can't delete Shared Drives root."));
+        return;
+    }
+
     qCDebug(GDRIVE) << "Deleting URL" << url << "- is it a file?" << isfile;
 
     const QUrlQuery urlQuery(url);
@@ -965,7 +1171,6 @@ void KIOGDrive::del(const QUrl &url, bool isfile)
         error(KIO::ERR_DOES_NOT_EXIST, url.path());
         return;
     }
-    const auto gdriveUrl = GDriveUrl(url);
     const QString accountId = gdriveUrl.account();
 
     // If user tries to delete the account folder, remove the account from the keychain
@@ -980,12 +1185,23 @@ void KIOGDrive::del(const QUrl &url, bool isfile)
         return;
     }
 
+    if (gdriveUrl.isSharedDrive()) {
+        qCDebug(GDRIVE) << "Deleting Shared Drive" << url;
+        if (deleteSharedDrive(url)) {
+            // error( will have been called in case of error
+            finished();
+        }
+        return;
+    }
+
     // GDrive allows us to delete entire directory even when it's not empty,
     // so we need to emulate the normal behavior ourselves by checking number of
     // child references
     if (!isfile) {
         ChildReferenceFetchJob referencesFetch(fileId, getAccount(accountId));
-        runJob(referencesFetch, url, accountId);
+        if (!runJob(referencesFetch, url, accountId)) {
+            return;
+        }
         const bool isEmpty = !referencesFetch.items().count();
 
         if (!isEmpty && metaData(QStringLiteral("recurse")) != QLatin1String("true")) {
@@ -995,12 +1211,10 @@ void KIOGDrive::del(const QUrl &url, bool isfile)
     }
 
     FileTrashJob trashJob(fileId, getAccount(accountId));
-    runJob(trashJob, url, accountId);
-
-    m_cache.removePath(url.path());
-
-    finished();
-
+    if (runJob(trashJob, url, accountId)) {
+        m_cache.removePath(url.path());
+        finished();
+    }
 }
 
 void KIOGDrive::rename(const QUrl &src, const QUrl &dest, KIO::JobFlags flags)
@@ -1039,9 +1253,26 @@ void KIOGDrive::rename(const QUrl &src, const QUrl &dest, KIO::JobFlags flags)
         return;
     }
 
+    if (srcGDriveUrl.isSharedDrive()) {
+        qCDebug(GDRIVE) << "Renaming Shared Drive" << srcGDriveUrl.filename() << "to" << destGDriveUrl.filename();
+        DrivesPtr drives = DrivesPtr::create();
+        drives->setId(sourceFileId);
+        drives->setName(destGDriveUrl.filename());
+
+        DrivesModifyJob modifyJob(drives, getAccount(sourceAccountId));
+        if (!runJob(modifyJob, src, sourceAccountId)) {
+            return;
+        }
+
+        finished();
+        return;
+    }
+
     // We need to fetch ALL, so that we can do update later
     FileFetchJob sourceFileFetchJob(sourceFileId, getAccount(sourceAccountId));
-    runJob(sourceFileFetchJob, src, sourceAccountId);
+    if (!runJob(sourceFileFetchJob, src, sourceAccountId)) {
+        return;
+    }
 
     const ObjectsList objects = sourceFileFetchJob.items();
     if (objects.count() != 1) {
@@ -1058,8 +1289,6 @@ void KIOGDrive::rename(const QUrl &src, const QUrl &dest, KIO::JobFlags flags)
         error(KIO::ERR_ACCESS_DENIED, dest.fileName());
         return;
     }
-    const auto srcPathComps = srcGDriveUrl.pathComponents();
-    const auto destPathComps = destGDriveUrl.pathComponents();
     if (destGDriveUrl.isAccountRoot()) {
         // user is trying to move to root -> we are only renaming
     } else {
@@ -1089,17 +1318,15 @@ void KIOGDrive::rename(const QUrl &src, const QUrl &dest, KIO::JobFlags flags)
         parentReferences << ParentReferencePtr(new ParentReference(destDirId));
     }
 
-    const QString destFileName = destPathComps.last();
-
     FilePtr destFile(sourceFile);
-    destFile->setTitle(destFileName);
+    destFile->setTitle(destGDriveUrl.filename());
     destFile->setParents(parentReferences);
 
     FileModifyJob modifyJob(destFile, getAccount(sourceAccountId));
     modifyJob.setUpdateModifiedDate(true);
-    runJob(modifyJob, dest, sourceAccountId);
-
-    finished();
+    if (runJob(modifyJob, dest, sourceAccountId)) {
+        finished();
+    }
 }
 
 void KIOGDrive::mimetype(const QUrl &url)
@@ -1118,12 +1345,10 @@ void KIOGDrive::mimetype(const QUrl &url)
     const QString accountId = GDriveUrl(url).account();
 
     FileFetchJob fileFetchJob(fileId, getAccount(accountId));
-#if KGAPI_VERSION >= QT_VERSION_CHECK(5, 12, 0)
     fileFetchJob.setFields({File::Fields::Id, File::Fields::MimeType});
-#else
-    fileFetchJob.setFields(FileFetchJob::Id | FileFetchJob::MimeType);
-#endif
-    runJob(fileFetchJob, url, accountId);
+    if (!runJob(fileFetchJob, url, accountId)) {
+        return;
+    }
 
     const ObjectsList objects = fileFetchJob.items();
     if (objects.count() != 1) {
